class ArithmeticError:
    ""


class AssertionError:
    ""


class AttributeError:
    ""


class BaseException:
    ""


class EOFError:
    ""


Ellipsis = None


class Exception:
    ""


class GeneratorExit:
    ""


class ImportError:
    ""


class IndentationError:
    ""


class IndexError:
    ""


class KeyError:
    ""


class KeyboardInterrupt:
    ""


class LookupError:
    ""


class MemoryError:
    ""


class NameError:
    ""


class NotImplementedError:
    ""


class OSError:
    ""


class OverflowError:
    ""


class RuntimeError:
    ""


class StopIteration:
    ""


class SyntaxError:
    ""


class SystemExit:
    ""


class TypeError:
    ""


class UnicodeError:
    ""


class ValueError:
    ""


class ZeroDivisionError:
    ""


def abs():
    pass


def all():
    pass


def any():
    pass


def bin():
    pass


class bool:
    ""


class bytearray:
    ""

    def append():
        pass

    def extend():
        pass


class bytes:
    ""

    def count():
        pass

    def endswith():
        pass

    def find():
        pass

    def format():
        pass

    def index():
        pass

    def isalpha():
        pass

    def isdigit():
        pass

    def islower():
        pass

    def isspace():
        pass

    def isupper():
        pass

    def join():
        pass

    def lower():
        pass

    def lstrip():
        pass

    def replace():
        pass

    def rfind():
        pass

    def rindex():
        pass

    def rsplit():
        pass

    def rstrip():
        pass

    def split():
        pass

    def startswith():
        pass

    def strip():
        pass

    def upper():
        pass


def callable():
    pass


def chr():
    pass


class classmethod:
    ""


class dict:
    ""

    def clear():
        pass

    def copy():
        pass

    def fromkeys():
        pass

    def get():
        pass

    def items():
        pass

    def keys():
        pass

    def pop():
        pass

    def popitem():
        pass

    def setdefault():
        pass

    def update():
        pass

    def values():
        pass


def dir():
    pass


def divmod():
    pass


class enumerate:
    ""


def eval():
    pass


def exec():
    pass


class filter:
    ""


class float:
    ""


class frozenset:
    ""

    def add():
        pass

    def clear():
        pass

    def copy():
        pass

    def difference():
        pass

    def difference_update():
        pass

    def discard():
        pass

    def intersection():
        pass

    def intersection_update():
        pass

    def isdisjoint():
        pass

    def issubset():
        pass

    def issuperset():
        pass

    def pop():
        pass

    def remove():
        pass

    def symmetric_difference():
        pass

    def symmetric_difference_update():
        pass

    def union():
        pass

    def update():
        pass


def getattr():
    pass


def globals():
    pass


def hasattr():
    pass


def hash():
    pass


def help():
    pass


def hex():
    pass


def id():
    pass


def input():
    pass


class int:
    ""

    def from_bytes():
        pass

    def to_bytes():
        pass


def isinstance():
    pass


def issubclass():
    pass


def iter():
    pass


def len():
    pass


class list:
    ""

    def append():
        pass

    def clear():
        pass

    def copy():
        pass

    def count():
        pass

    def extend():
        pass

    def index():
        pass

    def insert():
        pass

    def pop():
        pass

    def remove():
        pass

    def reverse():
        pass

    def sort():
        pass


def locals():
    pass


class map:
    ""


def max():
    pass


def min():
    pass


def next():
    pass


class object:
    ""


def oct():
    pass


def open():
    pass


def ord():
    pass


def pow():
    pass


def print():
    pass


class range:
    ""


def repr():
    pass


class reversed:
    ""


def round():
    pass


class set:
    ""

    def add():
        pass

    def clear():
        pass

    def copy():
        pass

    def difference():
        pass

    def difference_update():
        pass

    def discard():
        pass

    def intersection():
        pass

    def intersection_update():
        pass

    def isdisjoint():
        pass

    def issubset():
        pass

    def issuperset():
        pass

    def pop():
        pass

    def remove():
        pass

    def symmetric_difference():
        pass

    def symmetric_difference_update():
        pass

    def union():
        pass

    def update():
        pass


def setattr():
    pass


class slice:
    ""


def sorted():
    pass


class staticmethod:
    ""


class str:
    ""

    def count():
        pass

    def endswith():
        pass

    def find():
        pass

    def format():
        pass

    def index():
        pass

    def isalpha():
        pass

    def isdigit():
        pass

    def islower():
        pass

    def isspace():
        pass

    def isupper():
        pass

    def join():
        pass

    def lower():
        pass

    def lstrip():
        pass

    def replace():
        pass

    def rfind():
        pass

    def rindex():
        pass

    def rsplit():
        pass

    def rstrip():
        pass

    def split():
        pass

    def startswith():
        pass

    def strip():
        pass

    def upper():
        pass


def sum():
    pass


class super:
    ""


class tuple:
    ""

    def count():
        pass

    def index():
        pass


class type:
    ""


class zip:
    ""
