AP_IF = 1
AUTH_MAX = 6
AUTH_OPEN = 0
AUTH_WEP = 1
AUTH_WPA2_PSK = 3
AUTH_WPA_PSK = 2
AUTH_WPA_WPA2_PSK = 4


def LAN():
    pass


MODE_11B = 1
MODE_11G = 2
MODE_11N = 4
PHY_LAN8720 = 0
PHY_TLK110 = 1
STA_IF = 0


def WLAN():
    pass


def phy_mode():
    pass
