class VfsFat:
    ""

    def chdir():
        pass

    def getcwd():
        pass

    def ilistdir():
        pass

    label = None

    def mkdir():
        pass

    def mkfs():
        pass

    def mount():
        pass

    def open():
        pass

    def remove():
        pass

    def rename():
        pass

    def rmdir():
        pass

    def stat():
        pass

    def statvfs():
        pass

    def umount():
        pass


def erase_filesystem():
    pass


def getmount():
    pass


def mount():
    pass


def remount():
    pass


def umount():
    pass
